package v2

import (
	"context"
	"fmt"

	uuid "github.com/satori/go.uuid"
	corev1 "k8s.io/api/core/v1"
	metav1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	quarkv1alpha1 "quark.netapp.io/volume-controller/pkg/apis/volumecontroller/v1alpha1"
	"quark.netapp.io/volume-controller/pkg/quarkconfig"
)

// NetAppClusterPeeringUpdate is used to signal the most recent state of a NetAppClusterPeering relation and the
// most recent changes.
type NetAppClusterPeeringUpdate struct {
	ClusterPeering    *quarkv1alpha1.NetAppClusterPeering
	UpdatedConditions []quarkv1alpha1.NetAppClusterPeeringCondition
}

// NetAppClusterPeeringDelete is used to signal the deletion of a NetAppClusterPeering relation
type NetAppClusterPeeringDelete struct {
	ClusterPeering *quarkv1alpha1.NetAppClusterPeering
}

func ConstructNetAppClusterPeering(
	ownerID string,
	clusterPeeringUUID string,
	storagePoolUUID, remoteStoragePoolUUID string,
	remoteNamespace, remoteLocation, remoteClusterName string,
) (clusterPeering *quarkv1alpha1.NetAppClusterPeering, err error) {
	// validate clusterPeeringUUID
	_, err = uuid.FromString(clusterPeeringUUID)
	if err != nil {
		return nil, fmt.Errorf("invalid value for clusterPeeringUUID")
	}

	remoteService := quarkconfig.GetClusterPeeringServiceName(remoteStoragePoolUUID)
	remoteEndpoint := remoteService + "." + remoteNamespace + ".svc.cluster.local"

	clusterPeering = &quarkv1alpha1.NetAppClusterPeering{
		TypeMeta: metav1.TypeMeta{},
		ObjectMeta: metav1.ObjectMeta{
			Name: clusterPeeringUUID,
			Labels: map[string]string{
				quarkconfig.OwnerIDLabel:           ownerID,
				quarkconfig.StoragePoolLabel:       storagePoolUUID,
				quarkconfig.RemoteStoragePoolLabel: remoteStoragePoolUUID,
				quarkconfig.RemoteLocationLabel:    remoteLocation,
				quarkconfig.RemoteClusterNameLabel: remoteClusterName,
			},
			Finalizers: []string{
				quarkconfig.ClusterPeeringFinalizer,
			},
		},
		Spec: quarkv1alpha1.NetAppClusterPeeringSpec{
			RemoteEndpoint: remoteEndpoint,
		},
	}
	return
}

func (client *NVCClient) NetAppClusterPeeringCreate(clusterPeering *quarkv1alpha1.NetAppClusterPeering) (*quarkv1alpha1.NetAppClusterPeering, error) {
	return client.quarkClient.QuarkV1alpha1().NetAppClusterPeerings(client.nvcNamespace).Create(
		context.TODO(), clusterPeering, metav1.CreateOptions{})
}

func (client *NVCClient) NetAppClusterPeeringDelete(clusterPeeringUUID string) error {
	return client.quarkClient.QuarkV1alpha1().NetAppClusterPeerings(client.nvcNamespace).Delete(
		context.TODO(), clusterPeeringUUID, metav1.DeleteOptions{})
}

func (client *NVCClient) NetAppClusterPeeringGet(clusterPeeringUUID string) (*quarkv1alpha1.NetAppClusterPeering, error) {
	return client.quarkClient.QuarkV1alpha1().NetAppClusterPeerings(client.nvcNamespace).Get(
		context.TODO(), clusterPeeringUUID, metav1.GetOptions{})
}

func (client *NVCClient) NetAppClusterPeeringGetAll() ([]*quarkv1alpha1.NetAppClusterPeering, error) {
	clusterPeeringsList, err := client.quarkClient.QuarkV1alpha1().NetAppClusterPeerings(client.nvcNamespace).List(
		context.TODO(), metav1.ListOptions{})
	if err != nil {
		return nil, err
	}
	clusterPeerings := make([]*quarkv1alpha1.NetAppClusterPeering, len(clusterPeeringsList.Items))
	for i := range clusterPeeringsList.Items {
		clusterPeerings[i] = &clusterPeeringsList.Items[i]
	}
	return clusterPeerings, nil
}

func NetAppClusterPeeringReady(clusterPeering *quarkv1alpha1.NetAppClusterPeering) bool {
	return (clusterPeering.Status.State == quarkv1alpha1.ClusterPeeringStatePeered)
}

func NetAppServiceMeshEnabled(clusterPeering *quarkv1alpha1.NetAppClusterPeering) bool {
	for _, cond := range clusterPeering.Status.Conditions {
		if cond.Type == quarkv1alpha1.ServiceMeshEnabled && cond.Status == corev1.ConditionTrue {
			return true
		}
	}
	return false
}
